﻿    #ifndef MOESSCALC_H
#define MOESSCALC_H
#define DegRad 0.01745329251994329576923690768489
#define RadDeg 57.295779513082320876798154814105

struct quadratic_params
{
double a, b, c;
};

/* Calculation of energy levels and transition propabilities:
    all values in mm/s!
*/
#include <QtGlobal>
#include <QMessageBox>
#include <QProgressDialog>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_multiroots.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_sf.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_const_mksa.h>
#include <gsl/gsl_randist.h>
#include <stdlib.h>
#include "plot2D.h"
#include <cmath>
#include <QTime>


//! parameters to execute GSL fitting of parameter distribution
typedef struct LeCearContainer{
    gsl_matrix *M;
    gsl_matrix *D0;
    gsl_matrix *D1;
    gsl_matrix *A;
    gsl_vector *S;
    int steps;
    double lambda;
    double count;
}LeCearContainer;

//! Cambridge algorithm state indentifier
enum alphamode{decrease_in,decrease_out,increase_out, increase_in,
               initialize, exceed_aim};

//! calculation of moessbauer spectra
/*! spectrum constructors of the form spectrum(I,cc,v,p) need the following
 *  arguments\n
 *  \param I: intensity, array of length cc for output
 *  \param cc: channel count
 *  \param v: velocity array of length cc
 *  \param: parameters array
 */
class MoessCalc{
    public:
        MoessCalc();
        MoessCalc(MoessCalc* mc);
        ~MoessCalc();
        //! sets teh properties of the nucleus, if no 57Fe desired
        void set_nucleus(double g12, double g32, double Eg, double M,
                         double Q);
        //! fill Hamiltonian, solve and precalculate transition propabilities
        void calc();
        //! get Transition Energy from initial state i to final State f
        double get_Egamma(int f, int i);
        //! transform V/Angstroem^2 into mm/s at axialsymmetric field gradient
        double Vzz_mms(double Vzz);
        //! get current quadrupol splitting in mm/s
        double get_dv_Douplett();
        //! Transition propability from i to f in crystal geometry
        /*! \param beta: polar angle of gamma beam to field gradient z-axis
         * \param gamma: azimutal angle of gamma beam to field gradient z-axis
         * \param f: final state index
         * \param i: initial state index
         */
        double I(double beta, double gamma, int f, int i);
        //! Transition propability of polarized radiation in crystal geometry
        /*! \param beta: polar angle of gamma beam to field gradient z-axis
         * \param gamma: azimutal angle of gamma beam to field gradient z-axis
         * \param alpha: azimutal angle between source and sample z-axis
         * \param f: final state index
         * \param i: initial state index
         * \param Int: intensity output
         * \param phi: tilting of the polarisation ellipse with respect to z-axis
         * \param xi: eccentricity of the polarization matrix
         */
        void I(double beta, double gamma, double alpha, int f, int i,
               double* Int, double* phi, double* xi);
        //! transition propabilities for a diagonal hamitonian
        /*! \param beta: incoming polar angle
         * \param f: finale state index
         * \param i: initial state index
         */
        double Idiag(double beta, int f, int i);
        //! Intensity estimation according to saturation for thick absorber
        double I(double ta);
        //! powder transition propabilities for the current hamiltonian
        double I(int f, int i);
        //! Lorentz functions I/(pi*dx)*1/(1+((x-x0)/dx)^2)
        double f_lorenz(double x, double dx,double x0,double I);
        //! Lorentz functions I*invdx/((1+((x-x0)*invdx)^2)*M_PI)
        double f_lorenz_fast(double x, double invdx,double x0,double I);
        //! single crystal spectrum, static Hamiltonian
        void get_SC_spectrum(double *I, int cc , double* v, double* p);
        //! single crystal spectrum, static Hamiltonian, but Vzz coordinates
        void get_SC_spectrum_G(double *I, int cc , double* v, double* p);
        //! single crystal spectrum, static Hamiltonian, but B coordinates
        void get_SC_spectrum_B(double *I, int cc , double* v, double* p);
        //! powder spectrum, static Hamiltonian
        void get_Powder_spectrum(double *I, int cc , double* v, double* p);
        //! powder spectrum, static Hamiltonian, levelled like thick absorber
        void get_Powder_spectrum_lev(double *I, int cc , double* v, double* p);
        //! static Hamiltonian with source in field
        void get_SrcFld_spectrum(double *I, int cc , double* v, double* p);
        //! ferromagnet in transverse field spectrum
        void get_fm_Powder_spectrum(double *I, int cc , double* v, double* p);
        //! antiferromagnet in transverse field spetrum
        void get_afm_Powder_spectrum(double *I, int cc , double* v, double* p);
        //! antiferromagnet in longitudinal field spetrum
        void get_afm_Powder_long_spectrum(double *I, int cc , double* v, double* p);
        void get_PseudoVoigt_profile(double *I, int cc , double* v, double* p);
        //! iron calibrations, recalculates sin-velocity using monitur voltage U
        void get_FeCal_spectrum(double *I, int cc , double* v, double* p,
                                double U);
        //! Blume-Jones model with arbitrary number of intermediate states
        /*! Physical Review 174.2 (1968): 351 */
        void get_Dyn_spectrum(double *I, int cc , double* v, double* p);
        //! phenomenological dyn spectrum with dirty beta-implementation
        void get_Dynbeta_spectrum(double *I, int cc , double* v, double* p);
        //! multi level relaxation model, asuming a potential barrier
        /*! Journal of Physics: Condensed Matter 23 (2011), Nr. 42, S.426003 */
        void get_MLR_spectrum(double *I, int cc , double* v, double* p);
        //! Fast Blume Jones for two states with diagonal Hamiltonian
        void get_DynDiagDiag_spectrum(double *I, int cc , double* v, double* p);
        //! Fast Blume Jones for three states with diagonal Hamiltonian
        void get_DynDiagDiagDiag_spectrum(double *I, int cc , double* v, double* p);
        //! adds gaussian noise to spectrum I with standard deviation sqrt(I)
        void add_noise(double *I, int cc);
        //! iron calibrations, recalculates triangular-velocity
        void get_FeCaltriang_spectrum(double *I, int cc , double* v, double* p,
                                double U);
        //! calculates FeAs spectrum (Europhys. Lett., 9 (l), pp. 87-92 (1989))
        void get_FeAs_spectrum(double *I, int cc , double* v, double* p);
        //! calculates Fe2As spectrum (JPSJ 48.6 (1980): 1937-1941.)
        void get_Fe2As_spectrum(double *I, int cc , double* v, double* p);
        //! Hesse Ruebartsch without smoothing parameter
        /*! a number of subspectra are weighted to fit the total spectrum in
         *  the best way using linear algebra in the same way as
         *  Hesse-Ruebartsch, but without smoothing
         * \param x: velocity data
         * \param y: intensity data
         * \param count: channel count of each spectrum
         * \param steps: subspectra count
         * \param Mspec: count*steps Matrix with subspectra
         * \param rho: weights of the subspectra
         */
        void SimpleInvert(double* x, double *y, int count, int steps,
                          double **Mspec, double* rho);
        //! determination of baseline and absorption area
        /*! simulates the spectrum with a high number of Lorentzian singlets
         *  using SimpleInvert function
         */
        double baseline(double* x,double* y,int count,
                      double* Baseline,double* Face,
                      double bperx, double omega,
                      double** rho,double** rhox, int *rhosteps, double* INew);
        //! simple baseline/face estimation just considering the spectrum border
        void baseline(double* x,double* y,int count,double* Baseline,
                      double* Face);
        //! Maximum Entroy method using Cambridge algorithm to track MEM path
        /*! ambridge algorithm: Mon. Not. R Astr. Soc. (1984) 211, 111-124\n
         *  MEM in Moessbauer: Brand, R. A., and G. Le Caër. "Improving the validity of Mössbauer hyperfine parameter distributions: the maximum entropy formalism and its applications."\n
         * \param M0: subspectra to be weighted
         * \param rho: weight histogram to be smoothed
         * \param rhoerr: errors of the weights
         * \param y: experimental spectrum
         * \param I0: base line
         * \param steps: subspectra counts
         * \param cc: channel count
         * \param stepsize: of algorithm
         * \param MaxLambda: maximum fitting weight
         * \param MaxDif: maximum accepted divergence off entropy gradient and
         * chi2 gradient
         */
        void MEM_Cambridge(double**M0, double* rho, double* rhoerr, double *y,
                           double I0, int steps, int cc, double stepsize,
                           int MaxIterations, double MaxLambda,
                           double MaxDif, double status[5]);
        //! Hesse Ruebartsch parameter distribution
        /*! (1974 J. Phys. E: Sci. Instrum. 7 526), uses Linear Algebra to
         *  weihght a number of subspectra to meet both best fitting and a
         *  smooth weight histogram
         * \param M0: Matrix of subspectra
         * \param rho: spectra weights to be find
         * \param y: data points
         * \param steps: subspectra count
         * \param cc: channel count of each subspectrum
         * \param smooth: smoothing parameter
         */
        void HesseRuebartsch(double**M0, double* rho, double *y,
                             int steps, int cc, double smooth);
        //! Hesse Ruebartsch, but restricted to positive weights
        /*! Journal of Physics E: Scientifc Instruments 12 (1979), Nr. 11, S. 1083 \n
         *  Instead of using linear algebra the determination of the weight
         *  histogram might be achieved by numeric minimization. For parameters
         *  see HesseRuebartsch
         */
        void LeCear(double**M0, double* rho, double *y,
                    int steps, int cc, double smooth);
        //! quadratic dopplereffect in Debye approximation
        double dv_Debye(double T, double TD, double Meff);
        //! Debye Waller factor in Debye approximation
        double Abs_Debye(double T, double TD, double Meff);
        //! superfluid density in s-wave symmetry
        /*! Supercond. Sci. Technol.21(2008) 082003 (3pp)\n
         * \param SF0: superfluid density at T=0
         * \param Tc: critical temperature
         * \param T: temperature
         * \param delta0: gap value
         */
        double sWave_SuperfluidDensity(double SF0, double Tc, double T,
                                       double delta0);
        //! Werthamer-Helfand-Honenberg (WHH) model for upper critial field
        /*! see Collings, EW: Applied Superconductivity, Metallurgy, and
         *  Physics of Titanium Alloys: Volume 1: Fundamentals. Bd. 1. Springer
         *  1986
         */
        double Hc2_WHH(double T, double Tc, double dHc2dT, double alpha,
                       double lambdaSO);
        //! order parameter temperature dependence
        /*! with two critical exponents and a gaussian distributed transition
         *  temperature:\n
         *  OP(T)=int(0,\infty)[Gauss(t,Tn,sigma)*(1-(T/t)^alpha)^beta]dt\n
         * \param T: temperature
         * \param Tn: transition temperature mean
         * \param sigma: standarrd deviation of the transition temperature
         * \param alpha: first critical exponent
         * \param beta: second critical exponent
         */
        double OrderParameterDistr(double T, double Tn, double sigma,
                                   double alpha, double beta);
        //! Gaussian error function f(x)=2/sqrt(pi)*int(0,x)[exp(-a^2)]da
        double errorfunction(double arg);
        //! Convolution of source line with absorber spectrum
        /*! acording to the transmission integral the final absorption spectrum
         *  is the convolution of the Lorentzian source line with the absorber
         *  spectrum. The absorber spectrum represents the crosssection for
         *  i.e. it appears in the exponent of an exponential function.\n
         *  A symmetric velocity specing considerably can reduce the
         *  calculation costs for the convolution process. However the function
         *  can deal with asymmetric spacing as well.\n
         * \param SHI: thin absorber spectrum
         * \param v: velocities
         * \param count: channel count
         * \param A: spectral area of the thin-absorber-spectrum
         * \param I0: baseline
         * \param ta: effective thickness
         * \param omegasrc: source line width
         * \param omegaabs: absorber line width
         * \param fr: resonant fraction, i.e. the part of the photon count rate
         *  which in principle is able to be resonantly absorbed, i.e. it's not
         *  background radiation, it hits the sample, and it respects both
         *  source and sample Debye Waller factor
         */
        void TransmissionIntegral(double *SHI,double *v, int count,
                                  double A, double I0, double ta,
                                  double omegasrc, double omegaabs, double fr);
        //! calculates tilting of antiferromagnetic (afm) moments in field
        /*! \param Bex: external field
         * \param BJ: afm coupling constant
         * \param BA: anisotropy field
         * \param zeta: tilting of anisotropy axis with respect to Bex, degree
         * \param theta1: polar angle of moment 1 to Bex, degree
         * \param theta2: polar angle of moment 2 to Bex, degree
         */
        void afm_flipflop(double Bex, double BJ, double BA,  double zeta,
                          double* theta1, double*theta2);
        //! see afm_fliflop, but without antiferromagnetic coupling
        void fm_flip(double Bex, double BA,  double zeta, double* theta1);
        //! Mean field model for ferromagnetism
        double Mmf(double B, double n, double J, double T, double Tc,
                   double g);
        //! double Bezier_y(double* X, double* Y, int N, double x);
        double Bezier_y(double *X, double *Y, int N, double x);
        double Bezier_y(double x, double x1, double y1, double x2, double y2,
                        double x3, double y3, double x4, double y4);
        //! Bezier curve parametrized
        void Bezier(double* X, double* Y, int N, double t, double* x, double* y);
        //! Skyrmion model
        void get_skyrmion(double *I, int cc , double* v, double* p);
        void skyrmion_theta(double *theta, double thetaV, double k, double xi,
                            double dtheta, int maxEZ, double ca);
        void correlation_theta(double *theta, double thetaV, double k, double xi,
                            double dtheta, int maxEZ, double ca);
    private:
        //! B, Vzz, theta, phi and eta are used to recalculate Hamiltonian
        void fill_V();
        void set_parameters(double B, double Vzz,double eta, double theta,
                            double phi);
        gsl_matrix_complex *V0;         //!< ground state Hamiltonian
        gsl_matrix_complex *V1;         //!< excited state Hamiltonian
        gsl_matrix_complex *C0;         //!< ground state mixing coefficients
        gsl_matrix_complex *C1;         //!< exited state mixing coefficients
        //! used tridiagonal property of M in linear equation MX=b to speed up
        /*! Journal of Physics: Condensed Matter 23 (2011), Nr. 42, S. 426003
         */
        void SolveTridiagSys(gsl_matrix_complex* M, gsl_vector_complex* X,
                             gsl_vector_complex* b);

        double B;                       //!< magnetic hyerfine field
        double Vzz;                     //!< field gradient
        double theta;                   //!< polar angle of B to Vzz
        double phi;                     //!< azimutal angle of B to Vzz
        double eta;                     //!< asymmetry parameter
        double I0;                      //!< baseline
        double E0[2],E1[4];             //!< energy levels
        //!< quadrupol interaction
        /*! set by set_parameters function only */
        double a;
        //!< Zeeman splitting excited state
        /*! set by set_parameters function only */
        double b;                       //!< Dipolwechselwirkung fuer I=3/2
        //!< Zeeman splitting ground state
        /*! set by set_parameters function only */
        double b0;
        int firstcalc;                  //!< control parameter at instance start

        void init();                    //!< sets all constants

        ///@{
        //! Physical constants, set in init()
        double c,mun,Egamma,Q,g12,g32,Runi,Erec,kbolz,hbar_c_over_Egamma;
        ///@}

        //! transmission Integral core function, see TransmissionIntegral()
        /*! executes the convolution, but does not create the final spectrum
         * \param OneMinusT: output, convoluted absorption
         */
        void T_sigma(double *SHI,double *v, int count,
                     double A, double I0, double* OneMinusT,
                     double ta, double omegasrc, double omegaabs);

        //! Help functions/variables for the Cambridge algorithm
        /*! Maximum Entropy Method (MEM) is treated with the Cambridge
         *  algorithm: 1974 J. Phys. E: Sci. Instrum. 7 526\n
        */
        ///@{
        int steps;                      //!< maximum steps of the algorithm
        int count;                      //!< subsectra count
        gsl_vector *dchi2;              //!< current derivative of chi2
        gsl_matrix *d2chi2;             //!< Hesse Matrix concerning chi2
        gsl_vector *dS;                 //!< current derivative of entropy
        gsl_matrix *d2S;                //!< current 2nd derivative of entropy
        gsl_vector *yS;                 //!< experimental y-data
        gsl_vector *chi;                //!< current deviation from yS
        gsl_vector *P;                  //!< current weights
        gsl_vector *vtemp;              //!< temporary vector
        gsl_matrix  *M;                 //!< subspectra
        double cangle;                  //!< angle between dS and dchi2
        double cdif;                    //!< difference of dS and dchi2 direction
        double lagrange;                //!< current lagrange parameter
        double Pstepmax;                //!< maximum change of weights P per step
        void MEM_init(double**M, double *y);
        void MEM_free();
        void MEM_calc_derivatives();    //!< updates dS, d2S, chi, dchi
        void MEM_calc_diff();           //!< updates cdif, cangle
        // now specific Cambridge algorithm content follows
        int ssEWc;                      //!< subspace eigenvalue count
        gsl_vector *e[4];               //!< subspace base vectors
        gsl_vector *eON[4];             //!< orthonormalized subspace vectors
        gsl_vector *ssEW;               //!< subspace eigenvalues
        gsl_matrix *g;                  //!< metric tensor
        gsl_matrix *ssg;                //!< subspace metric tensor
        gsl_matrix *ssEV;               //!< subspace eigenvectors
        gsl_matrix *ssM;                //!< M matrix in subspace representation
        gsl_matrix  *MEV;               //!< eigenvectors of subspace M
        gsl_vector *MEW;                //!< eigenvalues of supspace M
        void MEM_Cambridge_create_subspace();
        void MEM_Cambridge_free_subspace();
        void MEM_Cambridge_calc_step(); //!< solves Lagrange equation
        void MEM_error_analysis(double* rho, double *rhoerror, double lambda);
        double MEM_Cambridge_lamda();   //!< returns current lagrange parameter
        double MEM_Cambridge_chi2();    //!< returns current fitting
        //! fine tune rho with bracket algorithm to achieve MaxLambda
        void MEM_Cambridge_tune_lambda(double* rho, double MaxLambda,
                                       double status[5]);
        ///@}


        //! Regression to fit multidimensional paraboloid to point cloud UNUSED
        /*! \param r: double[dim][N], coordinates of the N points
         * \param y: data value of the N points
         * \param N: data point count
         * \param dim: dimension, coordinate count
         * \param result: minimum of the paraboloid, coordinate vector
         * \param errors: certainity of the result coordinates
         */
        void fitParaboloid(double** r, double* y, int N,int dim,
                           double* result, double* errors);

        //! help function of fitParaboloid()
        double sum_up(double **x, int N, int dim, int* indeces);

        //! calculated theta and phi for given Vii and m vetors
        void theta_phi(double mt, double mp, double Vzzt, double Vzzp,
                       double Vxxt, double Vxxp, double *theta, double *phi);


        ///@{
        //! GSL work space for Hamiltonian treatment
        gsl_eigen_hermv_workspace * ws1;
        gsl_vector *eval1;
        gsl_eigen_hermv_workspace * ws0;
        gsl_vector *eval0;
        ///@}

        ///@{
        //! mixing coefficients of the eigenstates
        gsl_complex c12m12[2];
        gsl_complex c12p12[2];
        gsl_complex c32m32[4];
        gsl_complex c32m12[4];
        gsl_complex c32p12[4];
        gsl_complex c32p32[4];
        ///@}

        ///@{
        //! buffer values for transition propability calculation
        gsl_complex Km[2][4];
        gsl_complex K0[2][4];
        gsl_complex Kp[2][4];
        ///@}
};
///@{
//! global help functions, used for GSL Minimization
double Hafm_f(const gsl_vector * x, void * params );
void Hafm_df(const gsl_vector * x, void * params, gsl_vector * g );
void Hafm_fdf(const gsl_vector * x, void * params,double * f, gsl_vector *g);
///@}

///@{
//! global help functions, used for GSL Minimization
double Hfm_f(const gsl_vector * x, void * params );
void Hfm_df(const gsl_vector * x, void * params, gsl_vector * g );
void Hfm_fdf(const gsl_vector * x, void * params,double * f, gsl_vector *g);
///@}

///@{
//! global help functions, used for GSL Minimization
double BJ(double J, double y);
///@}

class CosSmearing{
    public:
        CosSmearing(double r_src, double r_smpl, double r_det, double z_smlp,
                    double z_det);
        ~CosSmearing();
        void update(double r_src, double r_smpl, double r_det, double z_smlp,
               double z_det);           //!< recalc rho
        //! propability density of a velocity shift dv at a give velocity v0
        double rho(double dv, double v0);
        //! propability of interval ddv at position dv
        double P(double v01, double v02, double v0);
        //! folds a Mössbauer spectrum with the velocity distribution
        void smear(double *SHI,double *v, int count, double I0);
    private:
        //! propability density of a cos value i
        double rho(double i);
        int steps;                      //!< histogram points
        double r0;                      //!< source radius
        double r1;                      //!< sample radius
        double r2;                      //!< detector radius
        double z1;                      //!< source-sample distance
        double z2;                      //!< source-detector distance
        double* Rho;                    //!< propability distribution of theta
        double one_minus_cosmin;
        double* costheta;               //!< precalculated cosine of theta
        double* sintheta;               //!< precalculated sin of theta
        //! does gamma ray from source hit the sample?
        bool in_smpl(double r,int i,double phi);
        //! does gamma ray from source hit the detector?
        bool in_det(double r,int i,double phi);
};


#endif // MOESSCALC_H
