﻿#ifndef fitthread_H
#define fitthread_H
#include <gsl/gsl_vector.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_multiroots.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_cdf.h>
#include <QProgressDialog>
#include <QMessageBox>
#include <QThread>
#include <QMutex>
#include <QTime>
#include "array.h"
#include "math.h"

//! different modi of error calculations are available in fitthread
/*! uncorrelated: returns parameter step to affect a change in chi2 of +1
 *  covariant: returns the diagonal elements of inverted Hesse matrix
 *  minos: changes and refits model until chi2 increases by 1
 *  montecarlo: probes random parametersets around the fit minimum and fits
 *  the resulting propability density function (pbf) - projected on every
 *  single parameter - with a gaussian distribution and returns gaussian width
 */
enum ErrMod{uncorrelated,covariant,minos,montecarlo};
//! container to be used in GSL minimization functions
typedef struct fitcontent{
    double (*chi2)(void*);
    double (*runchi2)(void*,int);
    void* instance;
    int cpara;
    int* IsChanged;
    double* param;
    int runi;
}fitcontent;
//! container to be used in GSL minimization functions
typedef struct fitcontentGauss{
    double* x;
    double* y;
    double* sigma;
    int count;
}fitcontentGauss;
//! tells what kind of calculation is to be executed
enum threattyp {fitseriell,fitglobal,errorseriell,errorglobal,notvalid};
//! container to sort correlation matrix elements using qSort()
/*! an estimate of parameter correlation is calculated from the Hessian matrix:
 *  "correlation(i,j)"=Hij^2/(Hii*Hjj) within the constructor
 */
class CorrelationEle{
    public:
        /*!
         * \param ia: matrix index of parameter a
         * \param ib: matrix index of parameter b
         * \param vala: f_aa derivative
         * \param valb: f_bb derivative
         * \param valab: f_ab derivative
         * \param xa: a value
         * \param xb: b-value
         */
        CorrelationEle(int ia, int ib, double vala,
                       double valb, double valab,
                       double xa, double xb);
        ~CorrelationEle();
        bool operator<(const CorrelationEle o2)const;
        int a,b;
        double va,vb;
        double correlation;
};

//! thread class to perform fitting and error calculation tasks
/*! Moessfit uses multithreading to efficiently fit and error calculate the
 *  data which usually consists of several runs. Multithreading is organized
 *  in the following way:\n
 *  (1) seriell fitting: the model specified by the user uses run specific
 *  parameters only. Then every run can be calculated individually (=seriell),
 *  no data is shared. For every run a fitthread will be created. All fitthread
 *  can work paralell.\n
 *  (2) global fitting: only one fitthread is used, multithreading is not used
 *  by using several fitthreads, but within fitmodel using QtConcurrent to
 *  calculate the spectra individually, if the global model has to be updated\n
 *  \n
 *  A fitthread knows the data to manipulate (which is owned by a fitmodel) and
 *  a function to recalculate the related fitmodel and get an updated chi2
 *  value.
 */
class fitthread: public QThread{
Q_OBJECT
public:
    void run();
    void exit();
    //! global fitting
    /*! \param para: parameters to be changed
     *  \param changed: help array to indicate if a parameter was changed
     *  \param errors: intial step sizes
     *  \param cpara: parameter count
     *  \param instance: fitmodel instance, is owner and used to recalculate
     *  \param MaxIter: maximum iteration count
     *  \param conv: convergence criterium
     *  \param forceSimplex: use GSL Simplex minimizer instead internal one
     */
    fitthread(double* para, int* changed, double* errors, int cpara,
              double (*f)(void*), void* instance, int MaxIter,
              double conv, bool forceSimplex);
    //! seriell fitting
    /*! \param runi: run index in fitmodel\n
     *  for other parameters see first constructor
     */
    fitthread(double* para, int runi, double* errors, int cpara,
              double (*f)(void*,int), void* instance, int MaxIter,
              double conv);
    ~fitthread();
    //! global error calculation
    /*! \param MonteCarloIterations: sample count to be probed
     * \param crun: run count
     * \param cglobal: global parameter count\n
     * \param correlated:
     * for other parameters see first constructor
     */
    fitthread(double* para, int* changed, double *errors, int cpara,
              double (*chi2)(void*), void* instance, ErrMod mode,
              double (*fitting)(void*), int MonteCarloIterations,
              int crun, int cglobal, bool (*correlated)(void*,int,int));
    //! global error calculation
    fitthread(double* para, int runi, double *errors, int cpara,
              double (*chi2)(void*,int), void* instance, ErrMod mode,
              int MonteCarloIterations);
    //! see ErrMod
    void errors_covariant();
    //! see ErrMod
    /*! \param sigma: is the multiplier to uncorrelated error intervall. The
     *  region to probe is defined by these extended intervalls */
    void errors_MonteCarlo(double pdfmin, double sigmac);
    //! see ErrMod, EmitSignals may suppress Progressbar update at internal use
    void errors_uncorrelated(bool EmitSignals);
    //! see ErrMod
    void errors_MINOS();
public slots:
    void stop_calculations();
    //! fits a dim-dimensional paraboloid centered at zero at points x, UNUSED
    /*! x is points*dim array holding the arguments, y are the function values
     *  at x (a points-dimensional vetctor and fxy returns the elements of the
     *  Hesse Matrix H in the following order:
     *  fx = (f11, f12, f13 ..., f22, f23,...,fNN))
     *  The data can be used in a quadradic model: f(x) = 0.5*xHx
     */
    void paraboloid(double **x, double *y, int N, int points, double* fxy);
private:
    /*! \param x: data
     * \param y: data
     * \param sigma: data yerror
     * \param count: of data points
     * \param gsigma: initial line width and output
     * \param gy0: peak height
     * \param gx0: peak position
     */
    void fit_gaussian(double *x, double* y, double* sigma, int count,
                      double* gsigma, double* gy0, double* gx0);
    void fit_seriell();
    void fit_global();
    void fit_global_simplex();
    //! returns, whether two parameters influence the same run
    int correlation_possible(int xi, int yi);
    //! Hesse matrix estimation based on MonteCarlo method, UNUSED
    /*! a specific numer of random points in parameter space P will be probed to
     *  calculated chi2(P). The resulting point cloud will be fitted with a
     *  multidimensional parabolic model
     * \param H: Hesse matrix output
     * \param ss: stepssize
     * \param k: parameters
     */
    void get_Hesse(gsl_matrix *H, double* ss, double *k);
    //! calculates/estimates covariant errors from Hesse matrix
    /*! covariant errors are the diagonal elements of the covariance matrix.
     *  Im a linear model the inverse of the Hessian matrix is the covariance
     *  matrix. If the Hessian matrix is not positive definite, the Inverion
     *  fails. In that case, the offdiagonal elements are scaled down until
     *  positive definiteness is ahieved.
     * \param H: Hesse matrix
     * \param errors: error output
     * \param vals: parameter values
     * \param cpara: parameter count
     */
    void Hesse_to_covariant_errors(gsl_matrix *H, double* errors,
                                   double* vals, int cpara);
    //! generates a string with the 10 most highest correlated parameters
    QString correlations(gsl_matrix *H, int cpara, double* vals);
    //! calculates d^2/dx^2 chi2(v)
    /*! \param v: pointer to value to be modified
     * \param v0: parameter value at minimum
     * \param ss: step size pointer, updated so that chi2(v0+ss)-chi0=1 finally
     * \param chi0: minimum value
     * \return
     */
    double fxx(double* v, double v0, double* ss, double chi0);
    //! calculates d/dx chi2(v), UNUSED
    double fx(double* v, double v0, double* ss, double chi0);
    //! calculates d^2/dxdy chi2(vx,vy), see fitthread::fxx()
    double fxy(double* vx, double* vy,
               double v0x, double v0y,
               double fxx, double fyy,
               double ssx, double ssy, double chi0);
    double* para;                       //!< parameters to be modified
    //!< this thread works upon runi-th run of the parent fitmodel
    int runi;
    double* errors;                     //!< errors or steps size of parameter
    //! tells whether parameter was modified during the last iteration
    int* changed;
    int cpara;                          //!< parameter count
    int crun;                           //!< run count
    int cglobal;                        //!< global parameter count
    double (*f_global)(void*);          //!< returns globally updates chi2
    double (*fit)(void*);               //!< triggers fitting of the fitmodel
    double (*f_seriell)(void*,int);     //!< returns updated chi2 of single run
    //! tells, if 2 parameters can influence the same run
    bool (*correlated)(void*,int,int);
    //! returns f_global or f_seriell depending on fitthread::fitmodel
    double f();
    void* instance;                     //!< pointer to parent fitmodel
    int MaxIter;                        //!< maximum iterations
    threattyp fitmodel;
    int proceed_calc;                   //!< thread was yet not stopped
    double conv;                        //!< convergence criterium
    bool usesimplex;                    //!< instead of normal global algorithm
    int montecarloiterations;           //!< maximum probe count
    ErrMod errormode;                   //!< error calculation, see ErrMod
signals:
    //! progressbar in fitmodel should be updated
    void refresh_progress(int ID, int citer,double chi2);
    //! calculation is finished
    void finished(int ID);
    //! errors are updated and may be printed
    void errors_calculated();
    //! fitmodel instance should print error message
    void error_message(int runi, QString message);

};

#endif // fitthread_H
