﻿#include "plot2D.h"

plot2D::plot2D(QWidget *parent){
    setBackgroundRole(QPalette::Dark);
    setAutoFillBackground(true);
    setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
    setFocusPolicy(Qt::StrongFocus);
    setMouseTracking(true);
    xmin = -1;
    xmax = 1;
    ymin = -1;
    ymax = 1;
    margin = 5;
    ticklength = 5;
    linewidth = 1;
    dataarea.setRect(0,0,width(),height());
    legendarea.setRect(0,0,0,0);
    psets = lsets = 0;
    xpdata = xldata = ypdata = yldata = NULL;
    lcc = pcc = NULL;
    showzoombox = 0;
    showResiduum = 0;
    refresh_pixmap();
    freep = 0;
    freel = 0;
    showLegend = 0;
    wheelselection = -1;
}
void plot2D::zoom_out(){
    double x = xmax-xmin, y = ymax-ymin;
    xmin = xmin - x/2.;
    xmax = xmax + x/2.;
    ymin = ymin - y/2.;
    ymax = ymax + y/2.;
    refresh_pixmap();
}
void plot2D::zoom_in(){
    double x = xmax-xmin, y = ymax-ymin;
    xmin = xmin + x/4.;
    xmax = xmax - x/4.;
    ymin = ymin + y/4.;
    ymax = ymax - y/4.;
    refresh_pixmap();
}
QSize plot2D::sizeHint() const{
    return QSize(450,300);
}
QSize plot2D::minimumSizeHint() const{
    return QSize(300,200);
}
void plot2D::paintEvent(QPaintEvent *){
    QPainter p(this);
    p.drawPixmap(0,0,pm);
    p.setBrush(QBrush(QColor(0,255,0,100)));
    p.setPen(palette().light().color());
    if(showzoombox)
        p.drawRect(zoombox.normalized());
    if(legendarea.contains(mouseposition)){
        int howerheight = qMax(1,legendarea.height()/leglabels.length());
        int legendi = (mouseposition.y() - legendarea.top())
                     /howerheight;
        p.drawRect(legendarea.left(),
                   legendarea.top()+legendi*howerheight,
                   legendarea.width(),howerheight);
    }
    if( !xbarlabel.isNull() &&
       dataarea.contains(mouseposition)){
        double fontsize = qMax(6.,qMin(width(),height())/40.);
        p.setFont(QFont("arial",fontsize));
        QFontMetricsF q(p.font());
        QString mousestring = QString("%1, %2")
                .arg(xtoxdata(mouseposition.x()),0,'f',qMax(0.,-xprecission+3))
                .arg(ytoydata(mouseposition.y()),0,'f',qMax(0.,-yprecission+3));
        p.setPen(Qt::black);
        p.drawText(dataarea.right()-q.width(mousestring)-margin,
                   dataarea.bottom()-margin, mousestring);
    }
}
void plot2D::resizeEvent(QResizeEvent *){
    refresh_pixmap();
}
void plot2D::mousePressEvent(QMouseEvent *event){
    if(   event->button() == Qt::LeftButton
       && dataarea.contains(mouseposition)){
        showzoombox = 1;
        zoombox.setTopLeft(event->pos());
        zoombox.setBottomRight(event->pos());
    }
}
void plot2D::mouseMoveEvent(QMouseEvent *event){
    if(showzoombox)
        zoombox.setBottomRight(event->pos());
    mouseposition = event->pos();
    update();
}
void plot2D::mouseReleaseEvent(QMouseEvent *event){
    if(   event->button() == Qt::LeftButton &&
          zoombox.width() > 5 && zoombox.height() > 5 &&
          showzoombox){
        zoombox.setBottomRight(event->pos());
        if(zoombox.normalized().width() < 2 || zoombox.normalized().height()<2)
            return;
        double xmint = xtoxdata(zoombox.normalized().left());
        double xmaxt = xtoxdata(zoombox.normalized().right());
        double ymint = ytoydata(zoombox.normalized().bottom());
        double ymaxt = ytoydata(zoombox.normalized().top());
        xmin = xmint;
        ymin = ymint;
        xmax = xmaxt;
        ymax = ymaxt;
        refresh_pixmap();
    }
    if(mouseposition == event->pos() &&
       legendarea.contains(mouseposition)){
        int howerheight = legendarea.height()/leglabels.length();
        if(howerheight == 0)
            refresh_pixmap();
        else{
            int legendi = (mouseposition.y() - legendarea.top())
                         /howerheight;
            wheelselection = legendi;
            plot_single_set(legendi);
        }
    }
    showzoombox = 0;
}
void plot2D::keyPressEvent(QKeyEvent *event){
    switch(event->key()){
        case Qt::Key_Plus:
            zoom_in();
            break;
        case Qt::Key_Minus:
            zoom_out();
            break;
        default:
            emit keypressed(event);
    }
}
void plot2D::wheelEvent(QWheelEvent * event ){
    // execute the index change
    if (event->delta() < 0)
        wheelselection--;
    else
        wheelselection++;
    // may all data sets should be shown
    if(wheelselection == -1 || wheelselection == qMax(psets,lsets)){
        autoset_axis();
        return;
    }

    // transform back to normal indices of necessary
    if(wheelselection < 0)
        wheelselection = qMax(psets,lsets) + wheelselection + 1;
    if(wheelselection >= qMax(psets,lsets))
        wheelselection = wheelselection - qMax(psets,lsets) - 1;

    // now plot
    plot_single_set(wheelselection);
}
void plot2D::plot_current_set(){
    if((wheelselection < psets || wheelselection < lsets) &&
        wheelselection > -1)
        plot_single_set(wheelselection);
    else
        autoset_axis();
}

void plot2D::plot_single_set(int i){
    // save old pointers
    int oldpsets = psets;
    int oldlsets = lsets;
    int* oldpcc = pcc;
    int* olflcc = lcc;
    double** oldxpdata = xpdata;
    double** oldypdata = ypdata;
    double** oldxldata = xldata;
    double** oldyldata = yldata;
    QStringList oldlabels = leglabels;
    //overwrite pointers
    if(psets > i){
        psets = 1;
        xpdata = &(xpdata[i]);
        ypdata = &(ypdata[i]);
        pcc = &(pcc[i]);
    }
    if(lsets > i){
        lsets = 1;
        xldata = &(xldata[i]);
        yldata = &(yldata[i]);
        lcc = &(lcc[i]);
    }
    if(leglabels.length() > i)
        leglabels = QStringList(leglabels.at(i));

    // now paint
    autoset_axis();

    // restore pointers
     psets = oldpsets;
     lsets = oldlsets;
     pcc = oldpcc;
     lcc = olflcc;
     xpdata = oldxpdata;
     ypdata = oldypdata;
     xldata = oldxldata;
     yldata = oldyldata;
     leglabels = oldlabels;
}

void plot2D::refresh_pixmap(){
    pm = QPixmap(size());
    QPainter p(&pm);
    pm.fill();
    add_axis_labels(&p);
    add_grid(&p);
    p.setRenderHint(QPainter::Antialiasing);
    add_data(&p);
    add_residuum(&p);
    add_legend(&p);
    update();
}
double plot2D::xtoxdata(int x){
    return xmin + 1.*(x-dataarea.x())/dataarea.width()*(xmax-xmin);
}
double plot2D::ytoydata(int y){
    return ymax - 1.*(y-dataarea.y())/dataarea.height()*(ymax-ymin);
}
double plot2D::xdatatox(double x){
    return dataarea.left() + int((x-xmin)/(xmax-xmin)*dataarea.width());
}
double plot2D::ydatatoy(double y){
    return dataarea.bottom() - int((y-ymin)/(ymax-ymin)*dataarea.height());
}
double plot2D::resy(int i){
    return resheight*(3.-(yldata[0][i]-ypdata[0][i])
                          /sqrt(qMax(1.,ypdata[0][i])))/6.;
}
void plot2D::add_axis_labels(QPainter* p){
    double fontsize = qMax(6.,qMin(width(),height())/40.);
    QFont f("arial",fontsize);
    p->setFont(f);
    QFontMetricsF q(f);
    textheight = q.height();
    ticklength = textheight /2;
    f.setPixelSize(textheight);
    int daxmin = margin + ticklength, daxmax = this->width()-margin,
        daymin = margin             , daymax = this->height()-margin-ticklength
                                                       -margin-textheight;
    if(!title.isEmpty())
        daymin = daymin + textheight*1.5 + margin;
    if(showResiduum){
        resheight = this->height()/7;
        daymin = daymin + resheight;
    }
    if(!xbarlabel.isEmpty())
        daymax = daymax - margin - textheight;
    if(!ybarlabel.isEmpty())
        daxmin = daxmin + margin + textheight;

    /*y-Achsen Beschriftungsintervall festlegen*/
    ysteps = floor((daymax-daymin-margin)/(textheight + margin));
    double step = 1.*(ymax-ymin)/ysteps;
    yprecission = floor(log10(step));
    step = step/pow(10,yprecission+1);
    ystep = 0.1;
    if(step > 0.1)     ystep = 0.2;
    if(step > 0.2)     ystep = 0.25;
    if(step > 0.25)    ystep = 0.5;
    if(step > 0.5)     ystep = 1.;
    ystep = ystep * pow(10,yprecission+1);
    ystep0 = ystep*ceil(ymin/ystep);
    ysteps = floor((ymax-ystep0)/ystep);

    QString a = QString("%1").arg(ystep0,0,'f',qMax(0.,-yprecission));
    QString b = QString("%1").arg(ystep0+ysteps*ystep,0,'f',qMax(0.,-yprecission));

    daxmin =   daxmin
            + margin
            + qMax(q.width(a), q.width(b));

    /*x-Achsen Beschriftungsintervall festlegen*/
    double ww;
    int OK = 1, End = 0, tickcount = 0;
    while(OK){
        if(End){ tickcount--; OK = 0;}
        else     tickcount++;
        step = 1.*(xmax-xmin)/tickcount;
        xprecission = floor(log10(step));
        step = step/pow(10,xprecission+1);
        xstep = 0.1;
        if(step > 0.1)     xstep = 0.2;
        if(step > 0.2)     xstep = 0.25;
        if(step > 0.25)    xstep = 0.5;
        if(step > 0.5)     xstep = 1.;
        xstep = xstep * pow(10,xprecission+1);
        xstep0 = xstep*ceil(xmin/xstep);
        xsteps = qMax(1.,floor((xmax-xstep0)/xstep));
        ww = q.width(QString("-%1").arg(xstep0,0,'f',qMax(0.,-xprecission)));
        if( (daxmax-daxmin)/(xsteps) - margin
            < q.width(QString("-%1").arg(xstep0,0,'f',qMax(0.,-xprecission))))
            End = 1;
    }

    /*Zeichenbereich festetzen*/
    dataarea = QRect(daxmin,daymin,daxmax-daxmin,daymax-daymin);

    /*y-Achsen Zeichnen*/
    p->drawLine(dataarea.topLeft(),dataarea.bottomLeft());
    p->drawLine(dataarea.topRight(),dataarea.bottomRight());
    double y0;
    for(double y = ystep0; y < ymax; y = y + ystep){
        y0 = ydatatoy(y);
        p->drawLine(dataarea.left(),           y0,
                    dataarea.left()-ticklength,y0);
        p->drawText(0,y0-textheight/2,dataarea.left()-ticklength-margin,textheight,
                    Qt::AlignRight,QString("%1").arg(y,0,'f',qMax(0.,-yprecission)));
    }
    if(!ybarlabel.isNull()){
        p->save();
        //p->translate(100,100);
        p->rotate(-90);
        p->drawText(0,margin,-height(),textheight,Qt::AlignCenter,ybarlabel);
        p->restore();
    }


    /*x-Achsen Zeichnen*/
    p->drawLine(dataarea.bottomLeft(),dataarea.bottomRight());
    p->drawLine(dataarea.topLeft(),dataarea.topRight());
    double x0;
    for(double x = xstep0; x < xmax; x = x + xstep){
        x0 = xdatatox(x);
        p->drawLine(x0, dataarea.bottom(),
                    x0, dataarea.bottom()+ticklength);
        QString t = QString("%1").arg(x,0,'f',qMax(0.,-xprecission));
        if(q.width(t)/2+x0+margin < width())
            p->drawText(x0-1000,dataarea.bottom()+ticklength+margin,2001,
                        textheight,Qt::AlignCenter,t);
    }
    if(!xbarlabel.isNull())
        p->drawText(dataarea.left(),dataarea.bottom()+2*margin+ticklength+textheight,
                    dataarea.width(),textheight,Qt::AlignCenter,xbarlabel);

    /*Titel zeichnen*/
    if(!title.isNull()){
        f.setPixelSize(int(1.5*fontsize));
        p->setFont(f);
        p->drawText(0,margin,width(),textheight*1.5,Qt::AlignCenter,title);
    }
}
void plot2D::add_grid(QPainter* p){
    p->setPen(QColor(200,200,200));
    p->setClipRect(dataarea.adjusted(+1,+1,-1,-1));
    for(double y = ystep0; y < ymax; y = y + ystep){
        p->drawLine(dataarea.left() ,ydatatoy(y),
                    dataarea.right(),ydatatoy(y));
    }
    for(double x = xstep0; x < xmax; x = x + xstep){
        p->drawLine(xdatatox(x), dataarea.bottom(),
                    xdatatox(x), dataarea.top());
    }
    p->setPen(QColor(0,0,0));
}
void plot2D::add_data(QPainter* p){
    p->setClipRect(dataarea.adjusted(+1,+1,-1,-1));
    int maxcc = 0;
    for(int i = 0; i < lsets; i++) if(lcc[i] > maxcc) maxcc = lcc[i];

    /*Punkte malen*/
    double pointsize = 3;
    for(int y = 0; y < psets; y++){
        p->setPen(QPen(QBrush(colorgradient(y,0,psets-1)),linewidth));
        p->setBrush(QBrush(colorgradient(y,0,psets-1)));
        for(int i = 0; i < pcc[y]; i++){
            p->drawEllipse(xdatatox(xpdata[y][i])-pointsize/2.,
                           ydatatoy(ypdata[y][i])-pointsize/2.,
                           pointsize,pointsize);
        }

    }
    /*Linie malen*/
    QPointF* d = (QPointF*)malloc(maxcc*sizeof(QPointF));
    for(int y = 0; y < lsets; y++){
        p->setPen(QPen(QBrush(colorgradient(y,0,lsets-1)),linewidth));
        for(int i = 0; i < lcc[y]; i++){
            d[i].setX(xdatatox(xldata[y][i]));
            d[i].setY(ydatatoy(yldata[y][i]));
        }
        p->drawPolyline(d,lcc[y]);
    }
    free(d);
    p->setClipRect(this->rect());
}
void plot2D::add_residuum(QPainter* p){
    if(showResiduum &&
       psets != 0 && lsets != 0){
        /*Gitter malen*/
        QRect ResRect(dataarea.left(),dataarea.top()-resheight,
                      dataarea.width(),resheight);
        p->setClipRect(ResRect);

        //p->drawRect(ResRect);
        for(int i = 0; i < 7; i++){
            if(i == 3)  p->setPen(QColor(0,255,0));
            else        p->setPen(QColor(0,0,0));
            p->drawLine(ResRect.left(),ResRect.top()+1.*i/6*resheight,
                        ResRect.right(),ResRect.top()+1.*i/6*resheight);
        }
        /*Daten malen*/
        QPointF* d = (QPointF*)malloc(lcc[0]*sizeof(QPointF));
        p->setPen(QColor(255,0,0));
        for(int i = 0; i < qMin(lcc[0],pcc[0]); i++){
            d[i].setX(xdatatox(xldata[0][i]));
            d[i].setY(ResRect.top()+resy(i));
        }
        p->drawPolyline(d,qMin(lcc[0],pcc[0]));
        free(d);
        p->setClipRect(this->rect());
    }
}
void plot2D::add_legend(QPainter* p){
    if(showLegend && leglabels.length()>0 && dataarea.height()>10){
        double fontsize = qMax(6.,qMin(width(),height())/40.);
        QFontMetricsF q(QFont("arial",fontsize));
        while(leglabels.length()*q.height() > dataarea.height()-10){
            fontsize = fontsize * 0.9;
            q = QFontMetricsF(QFont("arial",fontsize));
        }
        int legendleft = dataarea.left() + 5;
        int legendbottom = dataarea.bottom() - 5;
        int legendright = legendleft;
        int legendtop = legendbottom-(leglabels.length()*q.height());
        p->setFont(QFont("arial",fontsize));
        for(int i = leglabels.length()-1; i >= 0; i--){
            p->setPen(colorgradient(1.*i,0,leglabels.length()-1));
            p->drawText(legendleft,
                        legendbottom-(leglabels.length()-i-1)*q.height(),
                        leglabels.at(i));
            legendright = qMax(legendright,int(ceil(legendleft + q.width(leglabels.at(i)))));
        }
        legendarea.setRect(legendleft,legendtop,legendright-legendleft,legendbottom-legendtop);
    }
}
void plot2D::set_labels(QString title, QString xbarlabel, QString ybarlabel){
    this->title = title;
    this->xbarlabel = xbarlabel;
    this->ybarlabel = ybarlabel;
}
void plot2D::set_linedata(int xcount, int ycount, double* xdata, double** ydata){
    set_legend(NULL);                 //delete Legend
    if(xcount == 0 || xdata == NULL || ycount == 0 || ydata == NULL){
        free_array(&xldata,lsets);
        free_array(&yldata,lsets);
        if(lcc != NULL){
            free(lcc);
            lcc = NULL;
        }
        lsets = 0;
    }
    else{
        double **xbuf = new double*[ycount];
        int* cbuf = new int[ycount];
        for(int i = 0; i < ycount; i++){
            xbuf[i] = xdata;
            cbuf[i] = xcount;
        }
        set_linedata(ycount,cbuf,xbuf,ydata);
        delete xbuf;
        delete cbuf;
    }
}
void plot2D::set_pointdata(int xcount, int ycount, double* xdata, double** ydata){
    set_legend(NULL);                 //delete Legend
    if(xcount == 0 || xdata == NULL || ycount == 0 || ydata == NULL){
        free_array(&xpdata,psets);
        free_array(&ypdata,psets);
        if(pcc != NULL){free(pcc); pcc = NULL;}
        psets = 0;
    }
    else{
        double **xbuf = new double*[ycount];
        int* cbuf = new int[ycount];
        for(int i = 0; i < ycount; i++){
            xbuf[i] = xdata;
            cbuf[i] = xcount;
        }
        set_pointdata(ycount,cbuf,xbuf,ydata);
        delete xbuf;
        delete cbuf;
    }
}
void plot2D::set_linedata(int cset, int* cc, double** xdata, double** ydata){
    set_legend(NULL);                 //delete Legend
    free_array(&xldata,lsets);
    free_array(&yldata,lsets);
    if(lcc != NULL){free(lcc);lcc=NULL;}
    if(cset == 0 || cc == NULL || xdata == NULL || ydata == NULL){
        xldata = yldata = NULL;
        lcc = NULL;
        lsets = 0;
    }
    else{
        alloc_array(&xldata,cset,cc);
        alloc_array(&yldata,cset,cc);
        lsets = cset;
        lcc = new int[cset];
        for(int i = 0; i < cset; i++){
            lcc[i] = cc[i];
            for(int k = 0; k < cc[i]; k++){
                xldata[i][k] = xdata[i][k];
                yldata[i][k] = ydata[i][k];
            }
        }
    }
}
void plot2D::set_pointdata(int cset, int* cc, double** xdata, double** ydata){
    set_legend(NULL);                 //delete Legend
    free_array(&xpdata,psets);
    free_array(&ypdata,psets);
    if(pcc != NULL){free(pcc);pcc=NULL;}
    if(cset == 0 || cc == NULL || xdata == NULL || ydata == NULL){
        xpdata = ypdata = NULL;
        pcc = NULL;
        psets = 0;
    }
    else{
        alloc_array(&xpdata,cset,cc);
        alloc_array(&ypdata,cset,cc);
        psets = cset;
        pcc = new int[cset];
        for(int i = 0; i < cset; i++){
            pcc[i] = cc[i];
            for(int k = 0; k < cc[i]; k++){
                xpdata[i][k] = xdata[i][k];
                ypdata[i][k] = ydata[i][k];
            }
        }
    }
}
QColor plot2D::colorgradient(double x,double xmin,double xmax){
    QColor c;
    double q = (xmax-x)/(xmax-xmin);
    if(q<0.00)  {c.setRgb(255,255,255);             return c;} //gruen
    if(q<0.25)  {c.setRgb(q/0.25*255,255,0);        return c;} //gruen->gelb
    if(q<0.50)  {c.setRgb(255,(0.5-q)/0.25*255,0);  return c;} //gelb->rot
    if(q<0.75)  {c.setRgb(255,0,(q-0.5)/0.25*255);  return c;} //rot->violett
    if(q<=1.00) {c.setRgb((1.0-q)/0.25*255,0,255);  return c;} //violett->blau
    else        {c.setRgb(0,0,0);                   return c;} //blau
}
void plot2D::autoset_axis(){
    if(psets == 0 && lsets == 0)
        ymin = ymax = xmin = xmax = 0;
    else{
        ymax = xmax = -1e300;
        ymin = xmin = +1e300;
    }
    for(int i = 0; i < psets; i++)
        for(int j = 0; j < pcc[i]; j++){
            if(xpdata[i][j] > xmax) xmax = xpdata[i][j];
            if(ypdata[i][j] > ymax) ymax = ypdata[i][j];
            if(xpdata[i][j] < xmin) xmin = xpdata[i][j];
            if(ypdata[i][j] < ymin) ymin = ypdata[i][j];
        }
    for(int i = 0; i < lsets; i++)
        for(int j = 0; j < lcc[i]; j++){
            if(xldata[i][j] > xmax) xmax = xldata[i][j];
            if(yldata[i][j] > ymax) ymax = yldata[i][j];
            if(xldata[i][j] < xmin) xmin = xldata[i][j];
            if(yldata[i][j] < ymin) ymin = yldata[i][j];
        }
    if(xmin == xmax){xmin = xmin - 1; xmax = xmax + 1;}
    if(ymin == ymax){ymin = ymin - 1; ymax = ymax + 1;}
    refresh_pixmap();
}
void plot2D::set_linewidth(double linewidth){
    this->linewidth = linewidth;
}
void plot2D::data_to_clipboard(){
    QString cbt("");
    /*Kopfzeile*/
    for(int i = 0; i < psets; i++)
        if(leglabels.length()>i)
            cbt.append(QString("%1-x_point\t%1-y_point\t").arg(leglabels[i]));
        else
            cbt.append(QString("x_point%1\ty_point%1\t").arg(i));
    for(int i = 0; i < lsets; i++)
        if(leglabels.length()>i)
            cbt.append(QString("%1-x_line\t%1-y_line\t").arg(leglabels[i]));
        else
            cbt.append(QString("x_point%1\ty_point%1\t").arg(i));
    cbt.append("\n");
    int imax = 0;
    for(int i = 0; i < psets; i++) if(pcc[i] > imax) imax = pcc[i];
    for(int i = 0; i < lsets; i++) if(lcc[i] > imax) imax = lcc[i];
    /*Daten*/
    for(int k = 0; k < imax; k++){
        /*Punktdaten anhaengen*/
        for(int i = 0; i < psets; i++)
            if(pcc[i] >= k)
                cbt.append(QString("%1\t%2\t").arg(xpdata[i][k])
                                              .arg(ypdata[i][k]));
            else
                cbt.append(QString("\t\t"));
        /*Liniendaten anhaengen*/
        for(int i = 0; i < lsets; i++)
            if(lcc[i] >= k)
                cbt.append(QString("%1\t%2\t").arg(xldata[i][k])
                                              .arg(yldata[i][k]));
            else
                cbt.append(QString("\t\t"));
        cbt.append("\n");
    }
    /*In Zwischenablage ablegen*/
    QApplication::clipboard()->setText(cbt);
}
void plot2D::update_linedata(double** ydata, int ycount){
    if(xldata != NULL){
        double* xbuf = xldata[0];
        int xcbuf = lcc[0];
        if(freel){free(lcc); free(xldata);}
        lcc = new int[ycount];
        xldata  = new double*[ycount];
        for(int i = 0; i < ycount; i++){
            xldata[i] = xbuf;
            pcc[i] = xcbuf;
        }
        lsets = ycount;
        yldata= ydata;
        freel = 1;
    }
}
void plot2D::set_range(double plotxmin, double plotxmax,
                       double plotymin, double plotymax){
    xmin = plotxmin;
    xmax = plotxmax;
    ymin = plotymin;
    ymax = plotymax;
    refresh_pixmap();
}
void plot2D::show_residuum(){
    showResiduum = 1;
}
void plot2D::set_legend(QStringList* labels){
    leglabels.clear();
    if(labels == NULL){
        /*delete legend*/
        showLegend = 0;
    }
    else{
        showLegend = 1;
        leglabels.append(*labels);
    }
}
unsigned char** plot2D::get_bits(){
    if(!pm.isNull()){
        QImage img = pm.toImage();
        QRgb px;
        int widthc = width(), heightc = height();
        unsigned char** result = new unsigned char*[heightc];
        for(int h = 0; h < heightc; h++){
            result[h] = new unsigned char[widthc*3];
            for(int w = 0; w < widthc; w++){
                px = img.pixel(w,h);
                result[h][w*3+0] = qRed(px);
                result[h][w*3+1] = qGreen(px);
                result[h][w*3+2] = qBlue(px);
            }
        }
        return result;
    }
    else
        return NULL;
}
